use v6;

unit module Base40Encoder;

my %BP1 = ('!'       => <!! ?!>,
           '?'       => <!? ??>,
           'tc'      => <!0 !J !K ?4 ?A>,
           'ucall'   => <!1 !N ?5 ?B ?X>,
           'newline' => <!2 !O ?J ?.>,
	   '_'       => <!3 ?K>,
	   '"'       => <!4 ?L>,
	   '/'       => <!5 ?M>,
	   ';'       => <!6 ?N>,
	   ':'       => <!7 ?O>,
	   '='       => <!8 ?P>,
	   '%'       => <!9 ?Q>,
	   '#'       => <!A ?R>,
	   '>'       => <!B ?S>,
	   '<'       => <!C ?T>,
	   '*'       => <!D ?U>,
	   ','       => <!E ?V>,
	   '$'       => <!F ?W>,
	   "'"       => <!G ?Y>,
	   'ucone'   => <!H !I ?Z ?0>,
 	   "\\"      => <!L ?1>,
  	   '^'       => <!M ?2>,
 	   '('       => <!P ?3>,
 	   ')'       => <!Q ?6>,
	   '['       => <!R ?7>,
	   ']'       => <!S ?8>,
	   '{'       => <!T ?9>,
	   '}'       => <!U ?C>,
	   '&'       => <!V ?D>,
	   '@'       => <!W ?E>,
	   '~'       => <!X ?F>,
	   '-'       => <!Y ?G>,
	   '«'       => <!Z ?H>,
   	   '»'       => <!. ?I>,
	   '|'       => ("! ", "? ")
	  );
	   
my %special = ("!", "?").Set;

## say %BP1.perl;
## say "";

my %BP1-rev; 
@(%BP1{$_}).map( -> $a { %BP1-rev{$a} = $_ } ) for %BP1.keys;

## %BP1.keys.map( -> $b { @(%BP1{$b}).map( -> $a { %BP1-rev{$a} = $b } ) });

## say %BP1-rev.sort.perl;

our sub encode ($string)
{
  my @all = $string.comb;

  my $return  = "";
  my $current = "";

  while @all
  {
    $current = @all.shift;
    # die unless $current ~~ <base40> || <BP1>;

    if $current eq any(%BP1.keys)
    {
      $return ~= @(%BP1{$current}).pick;
      # say "S: $current -> { $return.substr(*-2) }";
    }
    elsif $current ~~ any("A" .. "Z", "a" .. "z")
    {
      $current ~= @all.shift while @all.elems && @all[0] ~~ any("A" .. "Z", "a" .. "z");
      # say "L: $current";
      $return ~= string-wrap($current);
    }
    elsif $current eq "\n"
    {
      # say "*: $current";
      $return ~= @(%BP1<newline>).pick;
    }
    else
    {
      # say "*: $current";
      $return ~= $current;
    }
  }

  return $return;
}

our sub decode ($string)
{
  my @all = $string.comb;

  my $return  = "";
  my $current = "";

  while @all
  {
    $current = @all.shift;

    # die unless $current ~~ <base40>;

    if $current eq any(%special.keys)
    {
      $current ~= @all.shift;
      if $current eq any( @(%BP1<ucone>) )
      {
        $return ~= @all.shift;
      }
      elsif $current eq any( @(%BP1<ucall>) )
      {
        $return ~= @all.shift while @all.elems && @all[0].&is-uppercase;
      }
      elsif $current eq any( @(%BP1<tc>) )
      {
        $return ~= @all.shift;
        $return ~= @all.shift.lc while @all.elems && @all[0].&is-uppercase;
      }
      elsif $current eq any( @(%BP1<newline>) )
      {
        $return ~= "\n";
      }
      elsif %BP1-rev{$current}
      {
        $return ~= %BP1-rev{$current};
      }
      else
      {
        say "******** $current *******"; # or perhaps DIE???
      }
    }
    else
    {
      $return ~= $current.lc;
    }

  }
  return $return;
}

sub is-uppercase ($letter)
{
  # say "ISU: $letter";

  return $letter eq any("A" .. "Z");
}

sub all-uppercase ($word)
{
  return $word ~~ /^<[A .. Z]>+$/;
}

sub all-lowercase ($word)
{
  return $word ~~ /^<[a .. z]>+$/;
}

sub string-wrap ($string)
{
  # return $string.uc;

  # return @(%BP1<newline>).pick if $string eq "\n";

  return $string.uc if $string.&all-lowercase;
  return @(%BP1<ucall>).pick ~ $string if $string.&all-uppercase;
  return @(%BP1<tc>).pick ~ $string.uc if $string.substr(0,1).&is-uppercase && $string.substr(1).&all-lowercase;

  my $return;
  for $string.comb -> $letter
  {
    $return ~=  @(%BP1<ucone>).pick if $letter.&is-uppercase;
    $return ~= $letter.uc;
  }
  return $return;
}

